@@ -1,5 +1,4 @@
 /******************************************************************************
- * $Id$
  *
  * Project:  OpenGIS Simple Features Reference Implementation
  * Purpose:  Class for representing a whole feature, and layer schemas.
@@ -9,29 +8,14 @@
  * Copyright (c) 1999,  Les Technologies SoftMap Inc.
  * Copyright (c) 2008-2013, Even Rouault <even dot rouault at spatialys.com>
  *
- * Permission is hereby granted, free of charge, to any person obtaining a
- * copy of this software and associated documentation files (the "Software"),
- * to deal in the Software without restriction, including without limitation
- * the rights to use, copy, modify, merge, publish, distribute, sublicense,
- * and/or sell copies of the Software, and to permit persons to whom the
- * Software is furnished to do so, subject to the following conditions:
- *
- * The above copyright notice and this permission notice shall be included
- * in all copies or substantial portions of the Software.
- *
- * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
- * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
- * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
- * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
- * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
- * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
- * DEALINGS IN THE SOFTWARE.
+ * SPDX-License-Identifier: MIT
  ****************************************************************************/
 
 #ifndef OGR_FEATURE_H_INCLUDED
 #define OGR_FEATURE_H_INCLUDED
 
 #include "cpl_atomic_ops.h"
+#include "gdal_fwd.h"
 #include "ogr_featurestyle.h"
 #include "ogr_geometry.h"
 #include "ogr_geomcoordinateprecision.h"
@@ -49,32 +33,6 @@
  * Simple feature classes.
  */
 
-#ifndef DEFINE_OGRFeatureH
-/*! @cond Doxygen_Suppress */
-#define DEFINE_OGRFeatureH
-/*! @endcond */
-#ifdef DEBUG
-typedef struct OGRFieldDefnHS *OGRFieldDefnH;
-typedef struct OGRFeatureDefnHS *OGRFeatureDefnH;
-typedef struct OGRFeatureHS *OGRFeatureH;
-typedef struct OGRStyleTableHS *OGRStyleTableH;
-#else
-/** Opaque type for a field definition (OGRFieldDefn) */
-typedef void *OGRFieldDefnH;
-/** Opaque type for a feature definition (OGRFeatureDefn) */
-typedef void *OGRFeatureDefnH;
-/** Opaque type for a feature (OGRFeature) */
-typedef void *OGRFeatureH;
-/** Opaque type for a style table (OGRStyleTable) */
-typedef void *OGRStyleTableH;
-#endif
-/** Opaque type for a geometry field definition (OGRGeomFieldDefn) */
-typedef struct OGRGeomFieldDefnHS *OGRGeomFieldDefnH;
-
-/** Opaque type for a field domain definition (OGRFieldDomain) */
-typedef struct OGRFieldDomainHS *OGRFieldDomainH;
-#endif /* DEFINE_OGRFeatureH */
-
 class OGRStyleTable;
 
 /************************************************************************/
@@ -124,6 +82,9 @@ class CPL_DLL OGRFieldDefn
     int bNullable;
     int bUnique;
 
+    // Used by drivers (GPKG) to track generated fields
+    bool m_bGenerated = false;
+
     std::string m_osDomainName{};  // field domain name. Might be empty
 
     std::string m_osComment{};  // field comment. Might be empty
@@ -136,6 +97,12 @@ class CPL_DLL OGRFieldDefn
     explicit OGRFieldDefn(const OGRFieldDefn *);
     ~OGRFieldDefn();
 
+    // Copy constructor
+    OGRFieldDefn(const OGRFieldDefn &oOther);
+
+    // Copy assignment operator
+    OGRFieldDefn &operator=(const OGRFieldDefn &oOther);
+
     void SetName(const char *);
 
     const char *GetNameRef() const
@@ -157,6 +124,7 @@ class CPL_DLL OGRFieldDefn
 
     void SetType(OGRFieldType eTypeIn);
     static const char *GetFieldTypeName(OGRFieldType);
+    static OGRFieldType GetFieldTypeByName(const char *);
 
     OGRFieldSubType GetSubType() const
     {
@@ -165,6 +133,7 @@ class CPL_DLL OGRFieldDefn
 
     void SetSubType(OGRFieldSubType eSubTypeIn);
     static const char *GetFieldSubTypeName(OGRFieldSubType);
+    static OGRFieldSubType GetFieldSubTypeByName(const char *);
 
     OGRJustification GetJustify() const
     {
@@ -226,6 +195,29 @@ class CPL_DLL OGRFieldDefn
         return bUnique;
     }
 
+    /**
+     * @brief Return whether the field is a generated field.
+     *
+     * At time of writing, only the GeoPackage and PG drivers fill that information. Consequently,
+     * only a returned value equal to TRUE can be fully trusted.
+     * @return TRUE if the field is a generated field, FALSE otherwise.
+     * @since GDAL 3.11
+     */
+    bool IsGenerated() const
+    {
+        return m_bGenerated;
+    }
+
+    /**
+     * @brief SetGenerated set the field generated status.
+     * @param bGeneratedIn TRUE if the field is a generated field, FALSE otherwise.
+     * @since GDAL 3.11
+     */
+    void SetGenerated(bool bGeneratedIn)
+    {
+        m_bGenerated = bGeneratedIn;
+    }
+
     void SetUnique(int bUniqueIn);
 
     const std::string &GetDomainName() const
@@ -245,7 +237,6 @@ class CPL_DLL OGRFieldDefn
     int IsSame(const OGRFieldDefn *) const;
 
     /** Convert a OGRFieldDefn* to a OGRFieldDefnH.
-     * @since GDAL 2.3
      */
     static inline OGRFieldDefnH ToHandle(OGRFieldDefn *poFieldDefn)
     {
@@ -253,7 +244,6 @@ class CPL_DLL OGRFieldDefn
     }
 
     /** Convert a OGRFieldDefnH to a OGRFieldDefn*.
-     * @since GDAL 2.3
      */
     static inline OGRFieldDefn *FromHandle(OGRFieldDefnH hFieldDefn)
     {
@@ -294,9 +284,6 @@ class CPL_DLL OGRFieldDefn
     /*! @endcond */
 
     TemporaryUnsealer GetTemporaryUnsealer();
-
-  private:
-    CPL_DISALLOW_COPY_ASSIGN(OGRFieldDefn)
 };
 
 #ifdef GDAL_COMPILATION
@@ -339,7 +326,6 @@ inline OGRFieldDefn::TemporaryUnsealer whileUnsealing(OGRFieldDefn *object)
  * OGRLayer::AlterGeomFieldDefn() should be called on a new instance of
  * OGRFieldDefn, for drivers that support AlterFieldDefn().
  *
- * @since OGR 1.11
  */
 
 class CPL_DLL OGRGeomFieldDefn
@@ -364,6 +350,12 @@ class CPL_DLL OGRGeomFieldDefn
     explicit OGRGeomFieldDefn(const OGRGeomFieldDefn *);
     virtual ~OGRGeomFieldDefn();
 
+    // Copy constructor
+    OGRGeomFieldDefn(const OGRGeomFieldDefn &oOther);
+
+    // Copy assignment operator
+    OGRGeomFieldDefn &operator=(const OGRGeomFieldDefn &oOther);
+
     void SetName(const char *);
 
     const char *GetNameRef() const
@@ -408,7 +400,6 @@ class CPL_DLL OGRGeomFieldDefn
     int IsSame(const OGRGeomFieldDefn *) const;
 
     /** Convert a OGRGeomFieldDefn* to a OGRGeomFieldDefnH.
-     * @since GDAL 2.3
      */
     static inline OGRGeomFieldDefnH ToHandle(OGRGeomFieldDefn *poGeomFieldDefn)
     {
@@ -416,7 +407,6 @@ class CPL_DLL OGRGeomFieldDefn
     }
 
     /** Convert a OGRGeomFieldDefnH to a OGRGeomFieldDefn*.
-     * @since GDAL 2.3
      */
     static inline OGRGeomFieldDefn *FromHandle(OGRGeomFieldDefnH hGeomFieldDefn)
     {
@@ -457,9 +447,6 @@ class CPL_DLL OGRGeomFieldDefn
     /*! @endcond */
 
     TemporaryUnsealer GetTemporaryUnsealer();
-
-  private:
-    CPL_DISALLOW_COPY_ASSIGN(OGRGeomFieldDefn)
 };
 
 #ifdef GDAL_COMPILATION
@@ -498,7 +485,7 @@ whileUnsealing(OGRGeomFieldDefn *object)
  * potentially other metadata.
  *
  * It is essentially a collection of field descriptions (OGRFieldDefn class).
- * Starting with GDAL 1.11, in addition to attribute fields, it can also
+ * In addition to attribute fields, it can also
  * contain multiple geometry fields (OGRGeomFieldDefn class).
  *
  * It is reasonable for different translators to derive classes from
@@ -549,53 +536,53 @@ class CPL_DLL OGRFeatureDefn
      *
      * Note: fields should not be added or removed while iterating over them.
      */
-    struct CPL_DLL Fields
+    template <class OwnerT, class ChildT> struct CPL_DLL Fields
     {
       private:
-        OGRFeatureDefn *m_poFDefn;
+        OwnerT m_poFDefn;
 
       public:
-        inline explicit Fields(OGRFeatureDefn *poFDefn) : m_poFDefn(poFDefn)
+        inline explicit Fields(OwnerT poFDefn) : m_poFDefn(poFDefn)
         {
         }
 
-        struct CPL_DLL ConstIterator
+        struct CPL_DLL Iterator
         {
           private:
-            OGRFeatureDefn *m_poFDefn;
+            OwnerT m_poFDefn;
             int m_nIdx;
 
           public:
-            inline ConstIterator(OGRFeatureDefn *poFDefn, int nIdx)
+            inline Iterator(OwnerT poFDefn, int nIdx)
                 : m_poFDefn(poFDefn), m_nIdx(nIdx)
             {
             }
 
-            inline const OGRFieldDefn *operator*() const
+            inline ChildT operator*() const
             {
                 return m_poFDefn->GetFieldDefn(m_nIdx);
             }
 
-            inline ConstIterator &operator++()
+            inline Iterator &operator++()
             {
                 m_nIdx++;
                 return *this;
             }
 
-            inline bool operator!=(const ConstIterator &it) const
+            inline bool operator!=(const Iterator &it) const
             {
                 return m_nIdx != it.m_nIdx;
             }
         };
 
-        inline ConstIterator begin()
+        inline Iterator begin()
         {
-            return ConstIterator(m_poFDefn, 0);
+            return Iterator(m_poFDefn, 0);
         }
 
-        inline ConstIterator end()
+        inline Iterator end()
         {
-            return ConstIterator(m_poFDefn, m_poFDefn->GetFieldCount());
+            return Iterator(m_poFDefn, m_poFDefn->GetFieldCount());
         }
 
         inline size_t size() const
@@ -603,12 +590,7 @@ class CPL_DLL OGRFeatureDefn
             return static_cast<std::size_t>(m_poFDefn->GetFieldCount());
         }
 
-        inline OGRFieldDefn *operator[](size_t i)
-        {
-            return m_poFDefn->GetFieldDefn(static_cast<int>(i));
-        }
-
-        inline const OGRFieldDefn *operator[](size_t i) const
+        inline ChildT operator[](size_t i)
         {
             return m_poFDefn->GetFieldDefn(static_cast<int>(i));
         }
@@ -616,6 +598,9 @@ class CPL_DLL OGRFeatureDefn
 
     //! @endcond
 
+    /** Return type of GetFields() */
+    using NonConstFields = Fields<OGRFeatureDefn *, OGRFieldDefn *>;
+
     /** Return an object that can be used to iterate over non-geometry fields.
         \verbatim
         for( const auto* poFieldDefn: poFeatureDefn->GetFields() )
@@ -626,9 +611,27 @@ class CPL_DLL OGRFeatureDefn
 
         @since GDAL 3.7
      */
-    inline Fields GetFields()
+    inline NonConstFields GetFields()
     {
-        return Fields(this);
+        return NonConstFields(this);
+    }
+
+    /** Return type of GetFields() const */
+    using ConstFields = Fields<const OGRFeatureDefn *, const OGRFieldDefn *>;
+
+    /** Return an object that can be used to iterate over non-geometry fields.
+        \verbatim
+        for( const auto* poFieldDefn: poFeatureDefn->GetFields() )
+        {
+            // do something
+        }
+        \endverbatim
+
+        @since GDAL 3.12
+     */
+    inline ConstFields GetFields() const
+    {
+        return ConstFields(this);
     }
 
     //! @cond Doxygen_Suppress
@@ -658,8 +661,31 @@ class CPL_DLL OGRFeatureDefn
 
     virtual void AddFieldDefn(const OGRFieldDefn *);
     virtual OGRErr DeleteFieldDefn(int iField);
+
+    /**
+     * @brief StealFieldDefn takes ownership of the field definition at index detaching
+     *        it from the feature definition.
+     * This is an advanced method designed to be only used for driver implementations.
+     * @param iField index of the field definition to detach.
+     * @return a unique pointer to the detached field definition or nullptr if the index is out of range.
+     * @since GDAL 3.11
+     */
+    virtual std::unique_ptr<OGRFieldDefn> StealFieldDefn(int iField);
+
+    virtual void AddFieldDefn(std::unique_ptr<OGRFieldDefn> &&poFieldDefn);
+
     virtual OGRErr ReorderFieldDefns(const int *panMap);
 
+    /**
+     * @brief StealGeomFieldDefn takes ownership of the the geometry field definition at index
+     *        detaching it from the feature definition.
+     * This is an advanced method designed to be only used for driver implementations.
+     * @param iField index of the geometry field definition to detach.
+     * @return a unique pointer to the detached geometry field definition or nullptr if the index is out of range.
+     * @since GDAL 3.11
+     */
+    virtual std::unique_ptr<OGRGeomFieldDefn> StealGeomFieldDefn(int iField);
+
     virtual int GetGeomFieldCount() const;
     virtual OGRGeomFieldDefn *GetGeomFieldDefn(int i);
     virtual const OGRGeomFieldDefn *GetGeomFieldDefn(int i) const;
@@ -670,53 +696,53 @@ class CPL_DLL OGRFeatureDefn
      *
      * Note: fields should not be added or removed while iterating over them.
      */
-    struct CPL_DLL GeomFields
+    template <class OwnerT, class ChildT> struct CPL_DLL GeomFields
     {
       private:
-        OGRFeatureDefn *m_poFDefn;
+        OwnerT m_poFDefn;
 
       public:
-        inline explicit GeomFields(OGRFeatureDefn *poFDefn) : m_poFDefn(poFDefn)
+        inline explicit GeomFields(OwnerT poFDefn) : m_poFDefn(poFDefn)
         {
         }
 
-        struct CPL_DLL ConstIterator
+        struct CPL_DLL Iterator
         {
           private:
-            OGRFeatureDefn *m_poFDefn;
+            OwnerT m_poFDefn;
             int m_nIdx;
 
           public:
-            inline ConstIterator(OGRFeatureDefn *poFDefn, int nIdx)
+            inline Iterator(OwnerT poFDefn, int nIdx)
                 : m_poFDefn(poFDefn), m_nIdx(nIdx)
             {
             }
 
-            inline const OGRGeomFieldDefn *operator*() const
+            inline ChildT operator*() const
             {
                 return m_poFDefn->GetGeomFieldDefn(m_nIdx);
             }
 
-            inline ConstIterator &operator++()
+            inline Iterator &operator++()
             {
                 m_nIdx++;
                 return *this;
             }
 
-            inline bool operator!=(const ConstIterator &it) const
+            inline bool operator!=(const Iterator &it) const
             {
                 return m_nIdx != it.m_nIdx;
             }
         };
 
-        inline ConstIterator begin()
+        inline Iterator begin()
         {
-            return ConstIterator(m_poFDefn, 0);
+            return Iterator(m_poFDefn, 0);
         }
 
-        inline ConstIterator end()
+        inline Iterator end()
         {
-            return ConstIterator(m_poFDefn, m_poFDefn->GetGeomFieldCount());
+            return Iterator(m_poFDefn, m_poFDefn->GetGeomFieldCount());
         }
 
         inline size_t size() const
@@ -724,12 +750,7 @@ class CPL_DLL OGRFeatureDefn
             return static_cast<std::size_t>(m_poFDefn->GetGeomFieldCount());
         }
 
-        inline OGRGeomFieldDefn *operator[](size_t i)
-        {
-            return m_poFDefn->GetGeomFieldDefn(static_cast<int>(i));
-        }
-
-        inline const OGRGeomFieldDefn *operator[](size_t i) const
+        inline ChildT operator[](size_t i) const
         {
             return m_poFDefn->GetGeomFieldDefn(static_cast<int>(i));
         }
@@ -737,6 +758,9 @@ class CPL_DLL OGRFeatureDefn
 
     //! @endcond
 
+    /** Return type of GetGeomFields() */
+    using NonConstGeomFields = GeomFields<OGRFeatureDefn *, OGRGeomFieldDefn *>;
+
     /** Return an object that can be used to iterate over geometry fields.
         \verbatim
         for( const auto* poGeomFieldDefn: poFeatureDefn->GetGeomFields() )
@@ -747,9 +771,28 @@ class CPL_DLL OGRFeatureDefn
 
         @since GDAL 3.7
      */
-    inline GeomFields GetGeomFields()
+    inline NonConstGeomFields GetGeomFields()
     {
-        return GeomFields(this);
+        return NonConstGeomFields(this);
+    }
+
+    /** Return type of GetGeomFields() const */
+    using ConstGeomFields =
+        GeomFields<const OGRFeatureDefn *, const OGRGeomFieldDefn *>;
+
+    /** Return an object that can be used to iterate over geometry fields.
+        \verbatim
+        for( const auto* poGeomFieldDefn: poFeatureDefn->GetGeomFields() )
+        {
+            // do something
+        }
+        \endverbatim
+
+        @since GDAL 3.12
+     */
+    inline ConstGeomFields GetGeomFields() const
+    {
+        return ConstGeomFields(this);
     }
 
     virtual void AddGeomFieldDefn(const OGRGeomFieldDefn *);
@@ -804,7 +847,6 @@ class CPL_DLL OGRFeatureDefn
     static void DestroyFeatureDefn(OGRFeatureDefn *);
 
     /** Convert a OGRFeatureDefn* to a OGRFeatureDefnH.
-     * @since GDAL 2.3
      */
     static inline OGRFeatureDefnH ToHandle(OGRFeatureDefn *poFeatureDefn)
     {
@@ -812,7 +854,6 @@ class CPL_DLL OGRFeatureDefn
     }
 
     /** Convert a OGRFeatureDefnH to a OGRFeatureDefn*.
-     * @since GDAL 2.3
      */
     static inline OGRFeatureDefn *FromHandle(OGRFeatureDefnH hFeatureDefn)
     {
@@ -870,7 +911,7 @@ class CPL_DLL OGRFeatureDefn
  *
  * @param bSealFields Whether fields and geometry fields should be unsealed and
  *                    resealed.
- *                    This is generally desirabled, but in case of deferred
+ *                    This is generally desirable, but in case of deferred
  *                    resolution of them, this parameter should be set to false.
  * @since GDAL 3.9
  */
@@ -893,7 +934,7 @@ class CPL_DLL OGRFeature
 {
   private:
     GIntBig nFID;
-    OGRFeatureDefn *poDefn;
+    const OGRFeatureDefn *poDefn;
     OGRGeometry **papoGeometries;
     OGRField *pauFields;
     char *m_pszNativeData;
@@ -911,7 +952,7 @@ class CPL_DLL OGRFeature
     bool CopySelfTo(OGRFeature *poNew) const;
 
   public:
-    explicit OGRFeature(OGRFeatureDefn *);
+    explicit OGRFeature(const OGRFeatureDefn *);
     virtual ~OGRFeature();
 
     /** Field value. */
@@ -1152,15 +1193,14 @@ class CPL_DLL OGRFeature
      * (dereference) more than one iterator step at a time, since you will get
      * a reference to the same object (FieldValue) at each iteration step.
      *
-     * <pre>
+     * \code{.cpp}
      * for( auto&& oField: poFeature )
      * {
      *      std::cout << oField.GetIndex() << "," << oField.GetName()<< ": " <<
      * oField.GetAsString() << std::endl;
      * }
-     * </pre>
+     * \endcode
      *
-     * @since GDAL 2.3
      */
     ConstFieldIterator begin() const;
     /** Return end of field value iterator. */
@@ -1169,20 +1209,24 @@ class CPL_DLL OGRFeature
     const FieldValue operator[](int iField) const;
     FieldValue operator[](int iField);
 
+#if defined(__clang__)
+#pragma clang diagnostic push
+#pragma clang diagnostic ignored "-Wweak-vtables"
+#endif
+
     /** Exception raised by operator[](const char*) when a field is not found.
      */
-    class FieldNotFoundException : public std::exception
+    class FieldNotFoundException final : public std::exception
     {
     };
 
+#if defined(__clang__)
+#pragma clang diagnostic pop
+#endif
+
     const FieldValue operator[](const char *pszFieldName) const;
     FieldValue operator[](const char *pszFieldName);
 
-    OGRFeatureDefn *GetDefnRef()
-    {
-        return poDefn;
-    }
-
     const OGRFeatureDefn *GetDefnRef() const
     {
         return poDefn;
@@ -1194,6 +1238,7 @@ class CPL_DLL OGRFeature
 
     OGRErr SetGeometryDirectly(OGRGeometry *);
     OGRErr SetGeometry(const OGRGeometry *);
+    OGRErr SetGeometry(std::unique_ptr<OGRGeometry>);
     OGRGeometry *GetGeometryRef();
     const OGRGeometry *GetGeometryRef() const;
     OGRGeometry *StealGeometry() CPL_WARN_UNUSED_RESULT;
@@ -1203,11 +1248,6 @@ class CPL_DLL OGRFeature
         return poDefn->GetGeomFieldCount();
     }
 
-    OGRGeomFieldDefn *GetGeomFieldDefnRef(int iField)
-    {
-        return poDefn->GetGeomFieldDefn(iField);
-    }
-
     const OGRGeomFieldDefn *GetGeomFieldDefnRef(int iField) const
     {
         return poDefn->GetGeomFieldDefn(iField);
@@ -1225,6 +1265,7 @@ class CPL_DLL OGRFeature
     const OGRGeometry *GetGeomFieldRef(const char *pszFName) const;
     OGRErr SetGeomFieldDirectly(int iField, OGRGeometry *);
     OGRErr SetGeomField(int iField, const OGRGeometry *);
+    OGRErr SetGeomField(int iField, std::unique_ptr<OGRGeometry>);
 
     void Reset();
 
@@ -1241,11 +1282,6 @@ class CPL_DLL OGRFeature
         return poDefn->GetFieldDefn(iField);
     }
 
-    OGRFieldDefn *GetFieldDefnRef(int iField)
-    {
-        return poDefn->GetFieldDefn(iField);
-    }
-
     int GetFieldIndex(const char *pszName) const
     {
         return poDefn->GetFieldIndex(pszName);
@@ -1492,9 +1528,10 @@ class CPL_DLL OGRFeature
                          bool bUseISO8601ForDateTimeAsString = false);
 
     //! @cond Doxygen_Suppress
-    OGRErr RemapFields(OGRFeatureDefn *poNewDefn, const int *panRemapSource);
+    OGRErr RemapFields(const OGRFeatureDefn *poNewDefn,
+                       const int *panRemapSource);
     void AppendField();
-    OGRErr RemapGeomFields(OGRFeatureDefn *poNewDefn,
+    OGRErr RemapGeomFields(const OGRFeatureDefn *poNewDefn,
                            const int *panRemapSource);
     //! @endcond
 
@@ -1532,11 +1569,10 @@ class CPL_DLL OGRFeature
     void SetNativeData(const char *pszNativeData);
     void SetNativeMediaType(const char *pszNativeMediaType);
 
-    static OGRFeature *CreateFeature(OGRFeatureDefn *);
+    static OGRFeature *CreateFeature(const OGRFeatureDefn *);
     static void DestroyFeature(OGRFeature *);
 
     /** Convert a OGRFeature* to a OGRFeatureH.
-     * @since GDAL 2.3
      */
     static inline OGRFeatureH ToHandle(OGRFeature *poFeature)
     {
@@ -1544,7 +1580,6 @@ class CPL_DLL OGRFeature
     }
 
     /** Convert a OGRFeatureH to a OGRFeature*.
-     * @since GDAL 2.3
      */
     static inline OGRFeature *FromHandle(OGRFeatureH hFeature)
     {
@@ -1564,7 +1599,6 @@ struct CPL_DLL OGRFeatureUniquePtrDeleter
 //! @endcond
 
 /** Unique pointer type for OGRFeature.
- * @since GDAL 2.3
  */
 typedef std::unique_ptr<OGRFeature, OGRFeatureUniquePtrDeleter>
     OGRFeatureUniquePtr;
@@ -1645,7 +1679,7 @@ class CPL_DLL OGRFieldDomain
      *
      * This is the same as the C function OGR_FldDomain_Destroy().
      */
-    virtual ~OGRFieldDomain() = 0;
+    virtual ~OGRFieldDomain();
 
     /** Clone.
      *
@@ -1845,15 +1879,7 @@ class CPL_DLL OGRRangeFieldDomain final : public OGRFieldDomain
                         const OGRField &sMin, bool bMinIsInclusive,
                         const OGRField &sMax, bool bMaxIsInclusive);
 
-    OGRRangeFieldDomain *Clone() const override
-    {
-        auto poDomain = new OGRRangeFieldDomain(
-            m_osName, m_osDescription, m_eFieldType, m_eFieldSubType, m_sMin,
-            m_bMinIsInclusive, m_sMax, m_bMaxIsInclusive);
-        poDomain->SetMergePolicy(m_eMergePolicy);
-        poDomain->SetSplitPolicy(m_eSplitPolicy);
-        return poDomain;
-    }
+    OGRRangeFieldDomain *Clone() const override;
 
     /** Get the minimum value.
      *
@@ -1922,14 +1948,7 @@ class CPL_DLL OGRGlobFieldDomain final : public OGRFieldDomain
                        OGRFieldType eFieldType, OGRFieldSubType eFieldSubType,
                        const std::string &osBlob);
 
-    OGRGlobFieldDomain *Clone() const override
-    {
-        auto poDomain = new OGRGlobFieldDomain(
-            m_osName, m_osDescription, m_eFieldType, m_eFieldSubType, m_osGlob);
-        poDomain->SetMergePolicy(m_eMergePolicy);
-        poDomain->SetSplitPolicy(m_eSplitPolicy);
-        return poDomain;
-    }
+    OGRGlobFieldDomain *Clone() const override;
 
     /** Get the glob expression.
      *
@@ -1954,7 +1973,7 @@ struct swq_evaluation_context;
 class CPL_DLL OGRFeatureQuery
 {
   private:
-    OGRFeatureDefn *poTargetDefn;
+    const OGRFeatureDefn *poTargetDefn;
     void *pSWQExpr;
     swq_evaluation_context *m_psContext = nullptr;
 
@@ -1965,7 +1984,8 @@ class CPL_DLL OGRFeatureQuery
 
     static int CanUseIndex(const swq_expr_node *, OGRLayer *);
 
-    OGRErr Compile(OGRLayer *, OGRFeatureDefn *, const char *, int bCheck,
+    OGRErr Compile(const OGRLayer *, const OGRFeatureDefn *, const char *,
+                   int bCheck,
                    swq_custom_func_registrar *poCustomFuncRegistrar);
 
     CPL_DISALLOW_COPY_ASSIGN(OGRFeatureQuery)
@@ -1974,9 +1994,9 @@ class CPL_DLL OGRFeatureQuery
     OGRFeatureQuery();
     ~OGRFeatureQuery();
 
-    OGRErr Compile(OGRLayer *, const char *, int bCheck = TRUE,
+    OGRErr Compile(const OGRLayer *, const char *, int bCheck = TRUE,
                    swq_custom_func_registrar *poCustomFuncRegistrar = nullptr);
-    OGRErr Compile(OGRFeatureDefn *, const char *, int bCheck = TRUE,
+    OGRErr Compile(const OGRFeatureDefn *, const char *, int bCheck = TRUE,
                    swq_custom_func_registrar *poCustomFuncRegistrar = nullptr);
     int Evaluate(OGRFeature *);
 
